@@ -24,7 +24,7 @@ namespace ROCKSDB_NAMESPACE {
 // Test variations of WriteImpl.
 class DBWriteTest : public DBTestBase, public testing::WithParamInterface<int> {
  public:
-  DBWriteTest() : DBTestBase("/db_write_test", /*env_do_fsync=*/true) {}
+  DBWriteTest() : DBTestBase("db_write_test", /*env_do_fsync=*/true) {}
 
   Options GetOptions() { return DBTestBase::GetOptions(GetParam()); }
 
@@ -60,14 +60,15 @@ TEST_P(DBWriteTest, WriteStallRemoveNoSlowdownWrite) {
     std::string key = "foo" + std::to_string(a);
     WriteOptions wo;
     wo.no_slowdown = false;
-    dbfull()->Put(wo, key, "bar");
+    ASSERT_OK(dbfull()->Put(wo, key, "bar"));
   };
   std::function<void()> write_no_slowdown_func = [&]() {
     int a = thread_num.fetch_add(1);
     std::string key = "foo" + std::to_string(a);
     WriteOptions wo;
     wo.no_slowdown = true;
-    dbfull()->Put(wo, key, "bar");
+    Status s = dbfull()->Put(wo, key, "bar");
+    ASSERT_TRUE(s.ok() || s.IsIncomplete());
   };
   std::function<void(void*)> unblock_main_thread_func = [&](void*) {
     mutex.Lock();
@@ -77,13 +78,13 @@ TEST_P(DBWriteTest, WriteStallRemoveNoSlowdownWrite) {
   };
 
   // Create 3 L0 files and schedule 4th without waiting
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
 
   ROCKSDB_NAMESPACE::SyncPoint::GetInstance()->SetCallBack(
       "WriteThread::JoinBatchGroup:Start", unblock_main_thread_func);
@@ -104,7 +105,7 @@ TEST_P(DBWriteTest, WriteStallRemoveNoSlowdownWrite) {
   // write_thread
   FlushOptions fopt;
   fopt.wait = false;
-  dbfull()->Flush(fopt);
+  ASSERT_OK(dbfull()->Flush(fopt));
 
   // Create a mix of slowdown/no_slowdown write threads
   mutex.Lock();
@@ -145,7 +146,7 @@ TEST_P(DBWriteTest, WriteStallRemoveNoSlowdownWrite) {
   mutex.Unlock();
 
   TEST_SYNC_POINT("DBWriteTest::WriteStallRemoveNoSlowdownWrite:1");
-  dbfull()->TEST_WaitForFlushMemTable(nullptr);
+  ASSERT_OK(dbfull()->TEST_WaitForFlushMemTable(nullptr));
   // This would have triggered a write stall. Unblock the write group leader
   TEST_SYNC_POINT("DBWriteTest::WriteStallRemoveNoSlowdownWrite:2");
   // The leader is going to create missing newer links. When the leader
@@ -178,14 +179,15 @@ TEST_P(DBWriteTest, WriteThreadHangOnWriteStall) {
     std::string key = "foo" + std::to_string(a);
     WriteOptions wo;
     wo.no_slowdown = false;
-    dbfull()->Put(wo, key, "bar");
+    ASSERT_OK(dbfull()->Put(wo, key, "bar"));
   };
   std::function<void()> write_no_slowdown_func = [&]() {
     int a = thread_num.fetch_add(1);
     std::string key = "foo" + std::to_string(a);
     WriteOptions wo;
     wo.no_slowdown = true;
-    dbfull()->Put(wo, key, "bar");
+    Status s = dbfull()->Put(wo, key, "bar");
+    ASSERT_TRUE(s.ok() || s.IsIncomplete());
   };
   std::function<void(void *)> unblock_main_thread_func = [&](void *) {
     mutex.Lock();
@@ -195,13 +197,13 @@ TEST_P(DBWriteTest, WriteThreadHangOnWriteStall) {
   };
 
   // Create 3 L0 files and schedule 4th without waiting
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
-  Flush();
-  Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar");
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
+  ASSERT_OK(Flush());
+  ASSERT_OK(Put("foo" + std::to_string(thread_num.fetch_add(1)), "bar"));
 
   ROCKSDB_NAMESPACE::SyncPoint::GetInstance()->SetCallBack(
       "WriteThread::JoinBatchGroup:Start", unblock_main_thread_func);
@@ -222,7 +224,7 @@ TEST_P(DBWriteTest, WriteThreadHangOnWriteStall) {
   // write_thread
   FlushOptions fopt;
   fopt.wait = false;
-  dbfull()->Flush(fopt);
+  ASSERT_OK(dbfull()->Flush(fopt));
 
   // Create a mix of slowdown/no_slowdown write threads
   mutex.Lock();
@@ -243,7 +245,7 @@ TEST_P(DBWriteTest, WriteThreadHangOnWriteStall) {
   mutex.Unlock();
 
   TEST_SYNC_POINT("DBWriteTest::WriteThreadHangOnWriteStall:1");
-  dbfull()->TEST_WaitForFlushMemTable(nullptr);
+  ASSERT_OK(dbfull()->TEST_WaitForFlushMemTable(nullptr));
   // This would have triggered a write stall. Unblock the write group leader
   TEST_SYNC_POINT("DBWriteTest::WriteThreadHangOnWriteStall:2");
   // The leader is going to create missing newer links. When the leader finishes,
@@ -260,7 +262,7 @@ TEST_P(DBWriteTest, WriteThreadHangOnWriteStall) {
 TEST_P(DBWriteTest, IOErrorOnWALWritePropagateToWriteThreadFollower) {
   constexpr int kNumThreads = 5;
   std::unique_ptr<FaultInjectionTestEnv> mock_env(
-      new FaultInjectionTestEnv(Env::Default()));
+      new FaultInjectionTestEnv(env_));
   Options options = GetOptions();
   options.env = mock_env.get();
   Reopen(options);
@@ -287,7 +289,7 @@ TEST_P(DBWriteTest, IOErrorOnWALWritePropagateToWriteThreadFollower) {
     threads.push_back(port::Thread(
         [&](int index) {
           // All threads should fail.
-          auto res = Put("key" + ToString(index), "value");
+          auto res = Put("key" + std::to_string(index), "value");
           if (options.manual_wal_flush) {
             ASSERT_TRUE(res.ok());
             // we should see fs error when we do the flush
@@ -307,6 +309,11 @@ TEST_P(DBWriteTest, IOErrorOnWALWritePropagateToWriteThreadFollower) {
     threads[i].join();
   }
   ASSERT_EQ(1, leader_count);
+
+  // The Failed PUT operations can cause a BG error to be set.
+  // Mark it as Checked for the ASSERT_STATUS_CHECKED
+  dbfull()->Resume().PermitUncheckedError();
+
   // Close before mock_env destruct.
   Close();
 }
@@ -315,13 +322,13 @@ TEST_P(DBWriteTest, ManualWalFlushInEffect) {
   Options options = GetOptions();
   Reopen(options);
   // try the 1st WAL created during open
-  ASSERT_TRUE(Put("key" + ToString(0), "value").ok());
+  ASSERT_TRUE(Put("key" + std::to_string(0), "value").ok());
   ASSERT_TRUE(options.manual_wal_flush != dbfull()->TEST_WALBufferIsEmpty());
   ASSERT_TRUE(dbfull()->FlushWAL(false).ok());
   ASSERT_TRUE(dbfull()->TEST_WALBufferIsEmpty());
   // try the 2nd wal created during SwitchWAL
-  dbfull()->TEST_SwitchWAL();
-  ASSERT_TRUE(Put("key" + ToString(0), "value").ok());
+  ASSERT_OK(dbfull()->TEST_SwitchWAL());
+  ASSERT_TRUE(Put("key" + std::to_string(0), "value").ok());
   ASSERT_TRUE(options.manual_wal_flush != dbfull()->TEST_WALBufferIsEmpty());
   ASSERT_TRUE(dbfull()->FlushWAL(false).ok());
   ASSERT_TRUE(dbfull()->TEST_WALBufferIsEmpty());
@@ -329,7 +336,7 @@ TEST_P(DBWriteTest, ManualWalFlushInEffect) {
 
 TEST_P(DBWriteTest, IOErrorOnWALWriteTriggersReadOnlyMode) {
   std::unique_ptr<FaultInjectionTestEnv> mock_env(
-      new FaultInjectionTestEnv(Env::Default()));
+      new FaultInjectionTestEnv(env_));
   Options options = GetOptions();
   options.env = mock_env.get();
   Reopen(options);
@@ -337,7 +344,7 @@ TEST_P(DBWriteTest, IOErrorOnWALWriteTriggersReadOnlyMode) {
     // Forcibly fail WAL write for the first Put only. Subsequent Puts should
     // fail due to read-only mode
     mock_env->SetFilesystemActive(i != 0);
-    auto res = Put("key" + ToString(i), "value");
+    auto res = Put("key" + std::to_string(i), "value");
     // TSAN reports a false alarm for lock-order-inversion but Open and
     // FlushWAL are not run concurrently. Disabling this until TSAN is
     // fixed.
@@ -351,7 +358,9 @@ TEST_P(DBWriteTest, IOErrorOnWALWriteTriggersReadOnlyMode) {
     }
     */
     if (!options.manual_wal_flush) {
-      ASSERT_FALSE(res.ok());
+      ASSERT_NOK(res);
+    } else {
+      ASSERT_OK(res);
     }
   }
   // Close before mock_env destruct.
@@ -361,7 +370,7 @@ TEST_P(DBWriteTest, IOErrorOnWALWriteTriggersReadOnlyMode) {
 TEST_P(DBWriteTest, IOErrorOnSwitchMemtable) {
   Random rnd(301);
   std::unique_ptr<FaultInjectionTestEnv> mock_env(
-      new FaultInjectionTestEnv(Env::Default()));
+      new FaultInjectionTestEnv(env_));
   Options options = GetOptions();
   options.env = mock_env.get();
   options.writable_file_max_buffer_size = 4 * 1024 * 1024;
@@ -389,14 +398,14 @@ TEST_P(DBWriteTest, LockWalInEffect) {
   Options options = GetOptions();
   Reopen(options);
   // try the 1st WAL created during open
-  ASSERT_OK(Put("key" + ToString(0), "value"));
+  ASSERT_OK(Put("key" + std::to_string(0), "value"));
   ASSERT_TRUE(options.manual_wal_flush != dbfull()->TEST_WALBufferIsEmpty());
   ASSERT_OK(dbfull()->LockWAL());
   ASSERT_TRUE(dbfull()->TEST_WALBufferIsEmpty(false));
   ASSERT_OK(dbfull()->UnlockWAL());
   // try the 2nd wal created during SwitchWAL
-  dbfull()->TEST_SwitchWAL();
-  ASSERT_OK(Put("key" + ToString(0), "value"));
+  ASSERT_OK(dbfull()->TEST_SwitchWAL());
+  ASSERT_OK(Put("key" + std::to_string(0), "value"));
   ASSERT_TRUE(options.manual_wal_flush != dbfull()->TEST_WALBufferIsEmpty());
   ASSERT_OK(dbfull()->LockWAL());
   ASSERT_TRUE(dbfull()->TEST_WALBufferIsEmpty(false));
@@ -423,13 +432,14 @@ TEST_P(DBWriteTest, ConcurrentlyDisabledWAL) {
               ROCKSDB_NAMESPACE::WriteOptions write_option_default;
               std::string no_wal_key = no_wal_key_prefix + std::to_string(t) +
                                        "_" + std::to_string(i);
-              this->Put(no_wal_key, no_wal_value, write_option_disable);
+              ASSERT_OK(
+                  this->Put(no_wal_key, no_wal_value, write_option_disable));
               std::string wal_key =
                   wal_key_prefix + std::to_string(i) + "_" + std::to_string(i);
-              this->Put(wal_key, wal_value, write_option_default);
-              dbfull()->SyncWAL();
+              ASSERT_OK(this->Put(wal_key, wal_value, write_option_default));
+              ASSERT_OK(dbfull()->SyncWAL());
             }
-            return 0;
+            return;
         });
     }
     for (auto& t: threads) {
@@ -451,5 +461,6 @@ INSTANTIATE_TEST_CASE_P(DBWriteTestInstance, DBWriteTest,
 int main(int argc, char** argv) {
   ROCKSDB_NAMESPACE::port::InstallStackTraceHandler();
   ::testing::InitGoogleTest(&argc, argv);
+  RegisterCustomObjects(argc, argv);
   return RUN_ALL_TESTS();
 }
