@@ -19,6 +19,7 @@ int main() {
 #include <vector>
 
 #include "memory/arena.h"
+#include "rocksdb/db.h"
 #include "table/cuckoo/cuckoo_table_builder.h"
 #include "table/cuckoo/cuckoo_table_factory.h"
 #include "table/cuckoo/cuckoo_table_reader.h"
@@ -31,7 +32,6 @@ int main() {
 #include "util/string_util.h"
 
 using GFLAGS_NAMESPACE::ParseCommandLineFlags;
-using GFLAGS_NAMESPACE::SetUsageMessage;
 
 DEFINE_string(file_dir, "", "Directory where the files will be created"
     " for benchmark. Added for using tmpfs.");
@@ -69,7 +69,7 @@ class CuckooReaderTest : public testing::Test {
   CuckooReaderTest() {
     options.allow_mmap_reads = true;
     env = options.env;
-    env_options = EnvOptions(options);
+    file_options = FileOptions(options);
   }
 
   void SetUp(int num) {
@@ -89,12 +89,9 @@ class CuckooReaderTest : public testing::Test {
 
   void CreateCuckooFileAndCheckReader(
       const Comparator* ucomp = BytewiseComparator()) {
-    std::unique_ptr<WritableFile> writable_file;
-    ASSERT_OK(env->NewWritableFile(fname, &writable_file, env_options));
-    std::unique_ptr<WritableFileWriter> file_writer(new WritableFileWriter(
-        NewLegacyWritableFileWrapper(std::move(writable_file)), fname,
-        env_options));
-
+    std::unique_ptr<WritableFileWriter> file_writer;
+    ASSERT_OK(WritableFileWriter::Create(env->GetFileSystem(), fname,
+                                         file_options, &file_writer, nullptr));
     CuckooTableBuilder builder(
         file_writer.get(), 0.9, kNumHashFunc, 100, ucomp, 2, false, false,
         GetSliceHash, 0 /* column_family_id */, kDefaultColumnFamilyName);
@@ -110,12 +107,10 @@ class CuckooReaderTest : public testing::Test {
     ASSERT_OK(file_writer->Close());
 
     // Check reader now.
-    std::unique_ptr<RandomAccessFile> read_file;
-    ASSERT_OK(env->NewRandomAccessFile(fname, &read_file, env_options));
-    std::unique_ptr<RandomAccessFileReader> file_reader(
-        new RandomAccessFileReader(NewLegacyRandomAccessFileWrapper(read_file),
-                                   fname));
-    const ImmutableCFOptions ioptions(options);
+    std::unique_ptr<RandomAccessFileReader> file_reader;
+    ASSERT_OK(RandomAccessFileReader::Create(
+        env->GetFileSystem(), fname, file_options, &file_reader, nullptr));
+    const ImmutableOptions ioptions(options);
     CuckooTableReader reader(ioptions, std::move(file_reader), file_size, ucomp,
                              GetSliceHash);
     ASSERT_OK(reader.status());
@@ -140,12 +135,10 @@ class CuckooReaderTest : public testing::Test {
   }
 
   void CheckIterator(const Comparator* ucomp = BytewiseComparator()) {
-    std::unique_ptr<RandomAccessFile> read_file;
-    ASSERT_OK(env->NewRandomAccessFile(fname, &read_file, env_options));
-    std::unique_ptr<RandomAccessFileReader> file_reader(
-        new RandomAccessFileReader(NewLegacyRandomAccessFileWrapper(read_file),
-                                   fname));
-    const ImmutableCFOptions ioptions(options);
+    std::unique_ptr<RandomAccessFileReader> file_reader;
+    ASSERT_OK(RandomAccessFileReader::Create(
+        env->GetFileSystem(), fname, file_options, &file_reader, nullptr));
+    const ImmutableOptions ioptions(options);
     CuckooTableReader reader(ioptions, std::move(file_reader), file_size, ucomp,
                              GetSliceHash);
     ASSERT_OK(reader.status());
@@ -212,12 +205,12 @@ class CuckooReaderTest : public testing::Test {
   uint64_t file_size;
   Options options;
   Env* env;
-  EnvOptions env_options;
+  FileOptions file_options;
 };
 
 TEST_F(CuckooReaderTest, FileNotMmaped) {
   options.allow_mmap_reads = false;
-  ImmutableCFOptions ioptions(options);
+  ImmutableOptions ioptions(options);
   CuckooTableReader reader(ioptions, nullptr, 0, nullptr, nullptr);
   ASSERT_TRUE(reader.status().IsInvalidArgument());
   ASSERT_STREQ("File is not mmaped", reader.status().getState());
@@ -331,12 +324,12 @@ TEST_F(CuckooReaderTest, WhenKeyNotFound) {
   }
   auto* ucmp = BytewiseComparator();
   CreateCuckooFileAndCheckReader();
-  std::unique_ptr<RandomAccessFile> read_file;
-  ASSERT_OK(env->NewRandomAccessFile(fname, &read_file, env_options));
-  std::unique_ptr<RandomAccessFileReader> file_reader(
-      new RandomAccessFileReader(NewLegacyRandomAccessFileWrapper(read_file),
-                                 fname));
-  const ImmutableCFOptions ioptions(options);
+
+  std::unique_ptr<RandomAccessFileReader> file_reader;
+  ASSERT_OK(RandomAccessFileReader::Create(
+      env->GetFileSystem(), fname, file_options, &file_reader, nullptr));
+
+  const ImmutableOptions ioptions(options);
   CuckooTableReader reader(ioptions, std::move(file_reader), file_size, ucmp,
                            GetSliceHash);
   ASSERT_OK(reader.status());
@@ -407,7 +400,7 @@ std::string GetFileName(uint64_t num) {
     FLAGS_file_dir = test::TmpDir();
   }
   return test::PerThreadDBPath(FLAGS_file_dir, "cuckoo_read_benchmark") +
-         ToString(num / 1000000) + "Mkeys";
+         std::to_string(num / 1000000) + "Mkeys";
 }
 
 // Create last level file as we are interested in measuring performance of
@@ -416,15 +409,13 @@ void WriteFile(const std::vector<std::string>& keys,
     const uint64_t num, double hash_ratio) {
   Options options;
   options.allow_mmap_reads = true;
-  Env* env = options.env;
-  EnvOptions env_options = EnvOptions(options);
+  const auto& fs = options.env->GetFileSystem();
+  FileOptions file_options(options);
   std::string fname = GetFileName(num);
 
-  std::unique_ptr<WritableFile> writable_file;
-  ASSERT_OK(env->NewWritableFile(fname, &writable_file, env_options));
-  std::unique_ptr<WritableFileWriter> file_writer(new WritableFileWriter(
-      NewLegacyWritableFileWrapper(std::move(writable_file)), fname,
-      env_options));
+  std::unique_ptr<WritableFileWriter> file_writer;
+  ASSERT_OK(WritableFileWriter::Create(fs, fname, file_options, &file_writer,
+                                       nullptr));
   CuckooTableBuilder builder(
       file_writer.get(), hash_ratio, 64, 1000, test::Uint64Comparator(), 5,
       false, FLAGS_identity_as_first_hash, nullptr, 0 /* column_family_id */,
@@ -441,14 +432,13 @@ void WriteFile(const std::vector<std::string>& keys,
   ASSERT_OK(file_writer->Close());
 
   uint64_t file_size;
-  env->GetFileSize(fname, &file_size);
-  std::unique_ptr<RandomAccessFile> read_file;
-  ASSERT_OK(env->NewRandomAccessFile(fname, &read_file, env_options));
-  std::unique_ptr<RandomAccessFileReader> file_reader(
-      new RandomAccessFileReader(NewLegacyRandomAccessFileWrapper(read_file),
-                                 fname));
+  ASSERT_OK(
+      fs->GetFileSize(fname, file_options.io_options, &file_size, nullptr));
+  std::unique_ptr<RandomAccessFileReader> file_reader;
+  ASSERT_OK(RandomAccessFileReader::Create(fs, fname, file_options,
+                                           &file_reader, nullptr));
 
-  const ImmutableCFOptions ioptions(options);
+  const ImmutableOptions ioptions(options);
   CuckooTableReader reader(ioptions, std::move(file_reader), file_size,
                            test::Uint64Comparator(), nullptr);
   ASSERT_OK(reader.status());
@@ -470,18 +460,18 @@ void ReadKeys(uint64_t num, uint32_t batch_size) {
   Options options;
   options.allow_mmap_reads = true;
   Env* env = options.env;
-  EnvOptions env_options = EnvOptions(options);
+  const auto& fs = options.env->GetFileSystem();
+  FileOptions file_options(options);
   std::string fname = GetFileName(num);
 
   uint64_t file_size;
-  env->GetFileSize(fname, &file_size);
-  std::unique_ptr<RandomAccessFile> read_file;
-  ASSERT_OK(env->NewRandomAccessFile(fname, &read_file, env_options));
-  std::unique_ptr<RandomAccessFileReader> file_reader(
-      new RandomAccessFileReader(NewLegacyRandomAccessFileWrapper(read_file),
-                                 fname));
+  ASSERT_OK(
+      fs->GetFileSize(fname, file_options.io_options, &file_size, nullptr));
+  std::unique_ptr<RandomAccessFileReader> file_reader;
+  ASSERT_OK(RandomAccessFileReader::Create(fs, fname, file_options,
+                                           &file_reader, nullptr));
 
-  const ImmutableCFOptions ioptions(options);
+  const ImmutableOptions ioptions(options);
   CuckooTableReader reader(ioptions, std::move(file_reader), file_size,
                            test::Uint64Comparator(), nullptr);
   ASSERT_OK(reader.status());
